
import random
import pygame
from flow import colors, settings, utils

from flow.assets.sprite_loader import colorize, get_car, get_taillight, darken_and_gray
from flow.assets import sound_loader

# Map directions to their opposites
RIGHT_TURNS = {'W': 'N', 'E': 'S', 'N': 'E', 'S': 'W'}
LEFT_TURNS = {'N': 'W', 'S': 'E', 'E': 'N', 'W': 'S'}
OPPOSITES = {'N': 'S', 'S': 'N', 'E': 'W', 'W': 'E'}
ROTATIONS = {'N': 180, 'W': 270, 'E': 90, 'S': 0}


def ensure_axis(move, sign=1):
    x, y, r = move
    r = int(90 * round(float(r)/90))*sign
    return x, y, r


class BrakeLight(pygame.sprite.Sprite):
    def __init__(self, direction, x, y, width, height):
        super().__init__()
        self.rect = pygame.Rect(x, y, width, height)

        self.image = pygame.transform.scale(
            pygame.transform.rotate(
                get_taillight(),
                ROTATIONS[direction]+180
            ), (width, height)
        )


class Car(pygame.sprite.Sprite):
    """
    x, y should be the mid front of the car.
    """
    size = settings.CAR_SIZE
    color = colors.RED
    # speed = settings.CAR_SPEED

    car_image = get_car()

    def __init__(self, x=0, y=0, direction_of_travel=None, speed=None):
        super().__init__()

        self.x, self.y = x, y
        if speed:
            self.max_speed = speed
        else:
            self.max_speed = settings.CAR_SPEED
        self.speed = self.max_speed

        self.rect = pygame.Rect((self.x, self.y), (self.size, self.size))
        self.rotation = 0
        self.direction_of_travel = direction_of_travel
        self.original_image = colorize(
            get_car(),
            utils.random_color(min=50, max=180)
        )
        self.original_image = pygame.transform.scale(
            self.original_image,
            (self.size, self.size)
        )
        self.image = self.original_image
        self.set_rotation(ROTATIONS[direction_of_travel])

        self.current_intersection = None
        self.turn_manoeuvres = []
        self.waiting_for_light = False
        self.brakelight = None

    def dead(self):
        sound_loader.play_collision()
        self.image = darken_and_gray(self.image)

    def set_rotation(self, rotation):
        self.image = pygame.transform.rotate(
            self.original_image,
            rotation,
        )

        if rotation % 180:
            self.rect = pygame.Rect(self.x, self.y, self.size, self.size).inflate(0, -6)
        else:
            self.rect = pygame.Rect(self.x, self.y, self.size, self.size).inflate(-6, 0)
        self.rect.center = (self.x, self.y)

        self.rotation = rotation

    def set_pos(self, x, y):
        self.x = x
        self.y = y
        self.rect.center = (x, y)

    @property
    def pos(self):
        return (self.x, self.y)

    def check_brake_collisions(self, other_cars):
        for car in other_cars:
            if car.waiting_for_light:
                self.waiting_for_light = True
                return

    def check_collision(self, other_car):
        pass

    def check_traffic_light_collisions(self, lights):
        light_rects = []
        if self.rect is None:
            return
        for i in lights:
                light_rects.append(i.rect)
        collision = self.rect.collidelist(light_rects)
        if collision > -1:
            self.waiting_for_light = True
        else:
            self.waiting_for_light = False

    def make_turn(self, intersection):
        if self.current_intersection == intersection:
            return
        self.current_intersection = intersection
        direction = random.choice(['left', 'forward', 'right'])

        if direction == 'left':
            self.turn_manoeuvres = self.left_turn()
            self.direction_of_travel = LEFT_TURNS[self.direction_of_travel]
        elif direction == 'right':
            self.turn_manoeuvres = self.right_turn()
            self.direction_of_travel = RIGHT_TURNS[self.direction_of_travel]
        else:
            pass

    def delta(self):
        dx = 0
        dy = 0
        if self.direction_of_travel == 'N':
            dy = -1
        elif self.direction_of_travel == 'S':
            dy = 1
        elif self.direction_of_travel == 'E':
            dx = 1
        elif self.direction_of_travel == 'W':
            dx = -1
        return(dx, dy)

    def update(self):
        if self.waiting_for_light or self.speed < 1:
            return

        if not self.turn_manoeuvres:
            (dx, dy) = self.delta()
            self.set_pos(self.x + dx*self.speed,
                         self.y + dy*self.speed)
        else:
            (x, y, rotation) = self.turn_manoeuvres.pop(0)
            self.set_pos(x, y)
            self.set_rotation(rotation)

    def arrived(self):
        return self.x < -self.size or self.x > settings.WIDTH + self.size or \
           self.y < -self.size or self.y > settings.HEIGHT + self.size

    def left_turn(self):
        offset = settings.ROAD_BUFFER
        num_steps = settings.NUM_TURN_STEPS
        if self.direction_of_travel == 'N':
            destination = (
                self.x - self.size - offset,
                self.y - self.size - offset,
            )
        elif self.direction_of_travel == 'S':
            destination = (
                self.x + self.size + offset,
                self.y + self.size + offset,
            )
        elif self.direction_of_travel == 'E':
            destination = (
                self.x + self.size + offset,
                self.y - self.size - offset,
            )
        elif self.direction_of_travel == 'W':
            destination = (
                self.x - self.size - offset,
                self.y + self.size + offset
            )

        ox, oy = self.pos
        moves = []

        # First move me so my x/y is more forward
        first_steps = max((self.size) // (self.max_speed * 3), 1)
        total_steps = first_steps + num_steps
        dx, dy = self.delta()

        for i in range(1, first_steps+1):
            moves.append((
                ox + (self.speed // 3)*dx*i,
                oy + (self.speed // 3)*dy*i,
                self.rotation + i*(90/total_steps)
            ))

        start_turn_pos = moves[-1]
        ox, oy, orot = start_turn_pos
        ex, ey = destination

        for i in range(1, num_steps+1):
            moves.append((
                ox + int(((ex - ox) / num_steps)*i),
                oy + int(((ey - oy) / num_steps)*i),
                int(orot + i*(90/total_steps))
            ))

        return moves

    def right_turn(self):
        offset = settings.ROAD_BUFFER
        num_steps = settings.NUM_TURN_STEPS
        if self.direction_of_travel == 'N':
            destination = (
                self.x + self.size + offset,
                self.y - 2*self.size - offset,
            )
        elif self.direction_of_travel == 'S':
            destination = (
                self.x - self.size - offset,
                self.y + 2*self.size + offset
            )
        elif self.direction_of_travel == 'E':
            destination = (
                self.x + 2*self.size + offset,
                self.y + self.size + offset,
            )
        elif self.direction_of_travel == 'W':
            destination = (
                self.x - 2*self.size - offset,
                self.y - self.size - offset,
            )

        ox, oy = self.pos
        moves = []

        # First move me so my x/y is more forward
        first_steps = max((self.size*2) // (self.max_speed * 3), 1)
        total_steps = first_steps + num_steps
        dx, dy = self.delta()

        for i in range(1, first_steps+1):
            moves.append((
                ox + (self.speed // 3)*dx*i,
                oy + (self.speed // 3)*dy*i,
                self.rotation + i*(-90/total_steps)
            ))

        start_turn_pos = moves[-1]
        ox, oy, orot = start_turn_pos
        ex, ey = destination

        for i in range(1, num_steps+1):
            moves.append((
                ox + int(((ex - ox) / num_steps)*i),
                oy + int(((ey - oy) / num_steps)*i),
                int(orot + i*(-90/total_steps))
            ))
        moves[-1] = ensure_axis(moves[-1])
        return moves

    def go(self):
        self.brakelight = None
        self.speed = self.max_speed

    def stop(self):
        self.speed = 0
        # self.turn_manoeuvres = []

        if not self.brakelight:
            dist = settings.BRAKELIGHT_DIST
            if self.direction_of_travel == 'N':
                self.brakelight = BrakeLight('N', self.x - self.size//2 + 1,
                                             self.y + self.size//2, self.size, dist)
            if self.direction_of_travel == 'E':
                self.brakelight = BrakeLight('E', self.x - self.size//2 - dist,
                                             self.y - self.size//2 + 1, dist, self.size)
            if self.direction_of_travel == 'S':
                self.brakelight = BrakeLight('S', self.x - self.size//2 + 3,
                                             self.y - self.size//2 - dist, self.size, dist)
            if self.direction_of_travel == 'W':
                self.brakelight = BrakeLight('W', self.x + self.size//2,
                                             self.y - self.size//2 + 3, dist, self.size)

            sound_loader.play_horn()

    def kill(self):
        if self.brakelight:
            self.brakelight.kill()
        super().kill()
