import pygame
import random
import itertools

from flow import settings, colors, utils, events
from flow.car import Car
from flow.road import Road
from flow.intersection import Intersection
from flow.traffic_light import TrafficLight
from flow.signal import Signal
from flow.assets.sprite_loader import fill_space, get_fire

NEW_CAR_EVENT = pygame.USEREVENT + 1


def generate_gaps(v_roads, h_roads):
    """ Return a list of ((xmin, xmax), (ymin, ymax)) of free areas"""

    # Get a tuple of xmin, xmax
    x_areas = []
    last_pos = 0
    for x in v_roads:
        x_areas.append((last_pos, x))
        last_pos = x + settings.ROAD_WIDTH

    x_areas.append((last_pos, settings.WIDTH))
    # Get a tuple of ymin, ymax
    y_areas = []
    last_pos = 0
    for y in h_roads:
        y_areas.append((last_pos, y))
        last_pos = y + settings.ROAD_WIDTH
    y_areas.append((last_pos, settings.HEIGHT))
    return [a for a in itertools.product(x_areas, y_areas)]


class Map:
    def __init__(
        self,
        num_v_roads=1,
        num_h_roads=1,
        target_num_cars=10,
        collision_threshold=1,
        car_speed=5,
        rate_of_new_cars=5000,
        gap_factor=2,
    ):
        self.paused = False
        self.pass_fail = None
        self.target_num_cars = target_num_cars
        self.collision_threshold = collision_threshold
        self.car_speed = car_speed
        self.rate_of_new_cars = rate_of_new_cars
        self.num_cars_arrived = 0
        self.num_cars_crashed = 0
        self.font = pygame.font.SysFont('Comic Sans MS', 24)
        h_road_range = list(range(
            settings.ROAD_WIDTH * 2,
            settings.HEIGHT - (3 * settings.ROAD_WIDTH) + 1,
            settings.ROAD_WIDTH
        ))

        v_road_range = list(range(
            settings.ROAD_WIDTH * 2,
            settings.WIDTH - (3 * settings.ROAD_WIDTH) + 1,
            settings.ROAD_WIDTH
        ))

        v_roads = v_road_range[::gap_factor+1]
        h_roads = h_road_range[::gap_factor+1]

        while len(v_roads) > num_v_roads:
            v_roads.pop(random.randint(0, len(v_roads)-1))
        while len(h_roads) > num_h_roads:
            h_roads.pop(random.randint(0, len(h_roads)-1))

        self.roads = pygame.sprite.Group()
        for x in v_roads:
            self.roads.add(
                Road(x, orientation='vertical')
            )
        for y in h_roads:
            self.roads.add(
                Road(y, orientation='horizontal')
            )

        free_areas = generate_gaps(v_roads, h_roads)

        self.free_areas = []
        for (x1, x2), (y1, y2) in free_areas:
            self.free_areas.append(
                ((x1, y1), fill_space(x1, x2, y1, y2)))
        self.explosions = []
        self.car_collision_groups = []

        self.brakelights = pygame.sprite.Group()
        self.intersections = pygame.sprite.Group()
        self.traffic_lights = pygame.sprite.Group()
        for x in v_roads:
            for y in h_roads:
                self.intersections.add(
                    Intersection(x, y)
                )
                for direction in 'NSEW':
                    self.traffic_lights.add(
                        TrafficLight(x, y, direction)
                    )

        self.cars = pygame.sprite.Group()
        pygame.time.set_timer(events.NEW_CAR_EVENT, self.rate_of_new_cars)

    def add_cars_to_car_collision_group(self, cars):
        # Find the sprite group (if any):
        cars_set = set(cars)
        for group in self.car_collision_groups:
            if cars_set.intersection(set(group.sprites())):
                to_explode = cars_set.difference(set(group.sprites()))
                self.num_cars_crashed += len(to_explode)
                for car in to_explode:
                    self.explosions.append((car.rect.topleft, get_fire()))
                    car.dead()

                group.add(*to_explode)
                return
        group = pygame.sprite.Group()
        group.add(*cars)
        self.car_collision_groups.append(group)
        self.num_cars_crashed += len(cars)
        for car in cars:
            self.explosions.append((car.rect.topleft, get_fire()))
            car.dead()


    def mouse_pressed(self, pos):
        # create an on the fly class that has a rect attribute
        mouse_sprite = type(
            'ms',
            (object,),
            dict(rect=pygame.Rect(pos, (1, 1)))
        )
        # Check if clicked in any of the sprite groups of collided cars
        index = -1
        for i, group in enumerate(self.car_collision_groups):
            collision = pygame.sprite.spritecollide(mouse_sprite, group, False)
            if collision:
                for sprite in group.sprites():
                    sprite.kill()
                index = i
                break
        if index > -1:
            self.car_collision_groups.pop(index)

        # Change traffic lights
        collision = pygame.sprite.spritecollideany(
            mouse_sprite,
            self.traffic_lights,
            False
        )
        if collision:
            collision.mouse_pressed()

    def add_car(self, from_direction=None):
        entrypoints = []
        # get all entrypoints
        for road in self.roads.sprites():
            entrypoints.extend(road.entrypoints(Car.size))
        random.shuffle(entrypoints)

        # filter entrypoints if needed
        if from_direction:
            entrypoints = [
                (x, y, d) for x, y, d in entrypoints if d == from_direction
            ]

        # check if a shadow_car collides with anything, if no, create the car
        for x, y, direction in entrypoints:
            shadow_car = pygame.Rect(
                x,
                y,
                Car.size + settings.BRAKELIGHT_DIST,
                Car.size + settings.BRAKELIGHT_DIST
            )
            shadow_car.center = (x, y)
            collision = shadow_car.collidelist(
                [s.rect for s in self.cars.sprites()]
            )
            # something is occupying the space into which the new car would go
            if collision >= 0:
                continue
            else:
                self.cars.add(Car(x, y, direction, speed=self.car_speed))
                return
        print('No space for new car, ignoring NEW_CAR_EVENT')

    def draw(self, screen):
        for i in self.free_areas:
            screen.blit(i[1], i[0])
        self.roads.draw(screen)
        self.intersections.draw(screen)
        self.brakelights.draw(screen)
        self.cars.draw(screen)
        self.traffic_lights.draw(screen)
        s = pygame.Surface((180, 90))  # the size of your rect
        s.set_alpha(128)                # alpha level
        s.fill((255, 255, 255))
        screen.blit(s, (5, 5, 200, 100))
        for (pos, explosion) in self.explosions:
            explosion.blit(screen, pos)
        pos = pygame.Rect(10, 10, 300, 300)
        utils.drawText2(
            screen,
            settings.MD_FONT,
            colors.BLACK,
            "Arrived: {}/{}\nCrashed: {}/{}\n".format(
                self.num_cars_arrived,
                self.target_num_cars,
                self.num_cars_crashed,
                self.collision_threshold
            ),
            pos,
            aa=True,
        )

    def update(self):
        if self.paused:
            return
        cars_stopped = set()

        for (index, (pos, explosion)) in reversed(list(enumerate(self.explosions))):
            if explosion.isFinished():
                self.explosions.pop(index)

        # Check if any cars leave the screen
        for car in self.cars.sprites():
            if car.arrived():
                car.kill()
                self.num_cars_arrived += 1

        # Check for collision with other cars
        collisions = pygame.sprite.groupcollide(self.cars, self.cars, False, False)
        for (car, collided_with) in collisions.items():
            remaining_collisions = [c for c in collided_with if c is not car]
            if remaining_collisions:
                car.stop()
                self.brakelights.add(car.brakelight)
                cars_stopped.add(car)
                self.add_cars_to_car_collision_group(collided_with)

                #self.cars.remove(*collided_with)

        # Potentially turn at intersections
        collisions = pygame.sprite.groupcollide(self.cars, self.intersections, False, False)
        for (car, intersections) in collisions.items():
            car.make_turn(intersections[0])

        # Stop at red lights
        collisions = pygame.sprite.groupcollide(
            self.cars, self.traffic_lights, False, False)
        for (car, collided_with) in collisions.items():
            if collided_with and collided_with[0].is_red and \
               collided_with[0].stop_direction == car.direction_of_travel:
                car.stop()
                self.brakelights.add(car.brakelight)
                cars_stopped.add(car)

        # Stop at break lights
        collisions = pygame.sprite.groupcollide(
            self.cars, self.brakelights, False, False)
        for (car, collided_with) in collisions.items():
            if collided_with:
                car.stop()
                self.brakelights.add(car.brakelight)
                cars_stopped.add(car)

        for car in self.cars.sprites():
            if car not in cars_stopped:
                if car.brakelight:
                    self.brakelights.remove(car.brakelight)
                car.go()

        self.cars.update()

        if self.num_cars_crashed >= self.collision_threshold:
            self.finish("FAIL")
        elif self.num_cars_arrived >= self.target_num_cars:
            self.finish("PASS")

    def pause(self):
        self.paused = not self.paused
        if self.paused:
            print('PAUSED')
            pygame.time.set_timer(events.NEW_CAR_EVENT, 0)
        else:
            print('UNPAUSED')
            pygame.time.set_timer(events.NEW_CAR_EVENT, self.rate_of_new_cars)

    def finish(self, finish_state=None):
        if not self.pass_fail:
            pygame.time.set_timer(events.NEW_CAR_EVENT, 0)
            Signal().send_signal(
                "next",
                finish_state,
                {
                    'num_cars_arrived': self.num_cars_arrived,
                    'target_num_cars': self.target_num_cars,
                    'num_cars_crashed': self.num_cars_crashed,
                    'collision_threshold': self.collision_threshold,
                }
            )
        self.pass_fail = finish_state


# if __name__ == '__main__':
#     pygame.init()
#     clock = pygame.time.Clock()
#     screen = pygame.display.set_mode(settings.SCREEN_SIZE)
#     running = True
#
#     map = Map(6, 4, gap_factor=2)
#     pygame.time.set_timer(NEW_CAR_EVENT, settings.RATE_OF_NEW_CARS)
#
#     while running:
#         for event in pygame.event.get():
#             if event.type == pygame.QUIT:
#                 running = False
#             if event.type == NEW_CAR_EVENT:
#                 map.add_car()
#                 pygame.time.set_timer(NEW_CAR_EVENT, 0)
#
#             if event.type == pygame.MOUSEBUTTONUP:
#                 map.mouse_pressed(pygame.mouse.get_pos())
#
#
#
#
#         screen.fill(colors.WHITE)
#         map.update()
#         map.draw(screen)
#         pygame.display.flip()
#         clock.tick(settings.FPS)
