import random
import pygame
from flow import settings, colors, utils, mixins, difficulty
from flow.map import Map
from flow.signal import Signal


class Scene:
    font_color = colors.BLACK
    background_color = colors.BLACK
    fps = 60

    def __init__(self):
        pass

    def key_pressed(self, key):
        return False

    def mouse_pressed(self, pos):
        return False

    def draw(self, surface):
        pass

    def update(self):
        pass

    def is_finished(self):
        return False

    def finish(self):
        self.is_finished = True

    def write_text(self, surface, font=None):
        if not font:
            font = self.font

        text_box = (
            settings.WIDTH // 6,
            settings.HEIGHT // 10,
            settings.WIDTH - settings.WIDTH // 3,
            settings.HEIGHT - settings.HEIGHT // 5
        )
        utils.drawText2(
            surface,
            font,
            self.font_color,
            self.text,
            text_box,
        )


# class Button:
#     def __init__(self, text, x, y, color, on_press):
#         self.x, self.y = x, y
#         self.text = text
#         self.color = color
#         self.on_press = on_press
#         self.height = 60
#         self.width = 200
#         self.rect = pygame.Rect((x, y), (self.width, self.height))
#
#     def collide(self, pos):
#         if self.rect.collidepoint(pos):
#             self.on_press()
#             return True
#
#     def draw(self, surface):
#         pygame.draw.rect(surface, self.color, self.rect)
#         text = myfont.render(self.text, True, (0, 0, 0))
#         surface.blit(text, (self.x+self.width//2-text.get_width()//2,
#                             self.y+self.height//2-text.get_height()//2))


class Intro(mixins.TerminalTextMixin, Scene):
    font_color = colors.GREEN
    font = pygame.font.SysFont(settings.FONT, settings.FONT_SIZE_MD)
    text = utils.load_dialogue('intro')
    fps = 36

    def __init__(self):
        super().__init__()

    def draw(self, surface):
        self.flash(surface)

    def key_pressed(self, key):
        if key == pygame.K_ESCAPE:
            Signal().send_signal("escape")
            return True
        elif key == pygame.K_SPACE:
            Signal().send_signal("next")
            return True
        return False


class StatsPage(Scene):
    font_color = colors.GREEN
    text = utils.load_dialogue('stats')
    fps = 24
    font = settings.XS_FONT

    def __init__(self, *args):
        if args:
            self.finish_state = args[0]
            self.stats = args[1]
        else:
            self.finish_state = 'ARE A QUITTER!'
            self.stats = {
                'num_cars_arrived': 0,
                'target_num_cars': 0,
                'num_cars_crashed': 0,
                'collision_threshold': 0,
            }

    def key_pressed(self, key):
        if key == pygame.K_ESCAPE:
            Signal().send_signal("escape")
            return True
        elif key == pygame.K_SPACE:
            Signal().send_signal("next", self.finish_state)
            return True

    def mouse_pressed(self, pos):
        return False

    def write_text(self, surface):
        block1_text = '\n'.join(
            [t.format(finish_state=self.finish_state) for t in self.text[:2]]
        )

        block2_text = '\n'.join(
            [t.format(**self.stats) for t in self.text[2:4]]
        )

        to_draw = [{
            'rect': (
                settings.WIDTH // 6,
                settings.HEIGHT // 8,
                settings.WIDTH - settings.WIDTH // 3,
                settings.HEIGHT - settings.HEIGHT // 4
            ),
            'font': settings.MD_FONT,
            'text': block1_text,
        }, {
            'rect': (
                settings.WIDTH // 6,
                (settings.HEIGHT // 6)*2,
                settings.WIDTH - settings.WIDTH // 3,
                settings.HEIGHT // 3
            ),
            'font': settings.MD_FONT,
            'text': block2_text,
        }, {
            'rect': (
                settings.WIDTH // 6,
                (settings.HEIGHT // 8)*7,
                settings.WIDTH - settings.WIDTH // 3,
                settings.HEIGHT // 8
            ),
            'font': settings.SM_FONT,
            'text': '\n'.join(self.text[4:])
        }]

        for text in to_draw:
            utils.drawText2(
                surface,
                text['font'],
                self.font_color,
                text['text'],
                text['rect'],
            )

    def draw(self, surface):
        self.write_text(surface)

    def update(self):
        pass


class GameLevel(Scene):
    difficulty_settings = [
        difficulty.MaxOutDifficulty('num_v_roads', 2, max=4),
        difficulty.MaxOutDifficulty('num_h_roads', 2, max=4),
        difficulty.Difficulty('target_num_cars', 10, delta=5),
        difficulty.MinOutDifficulty('collision_threshold', 10, delta=-1, min=1),
        difficulty.MinOutDifficulty('rate_of_new_cars', 600, delta=-50, min=300),
        difficulty.MinOutDifficulty('gap_factor', 3, delta=-1, min=2)
    ]

    def __init__(self, *args):
        if args and args[0] == 'PASS':
            self.increase_difficulty()
        self.map = Map(**self.difficulty)

    @property
    def difficulty(self):
        return {d.name: d.value for d in self.difficulty_settings}

    def increase_difficulty(cls):
        print('increasing difficulty')
        difficulty_settings_idxs = list(range(len(cls.difficulty_settings)))
        increased = False
        while not increased:
            idx = random.choice(difficulty_settings_idxs)
            increased = cls.difficulty_settings[idx].increase()
            if not increased:
                idx_idx = difficulty_settings_idxs.index(idx)
                difficulty_settings_idxs.pop(idx_idx)

    def add_car(self):
        self.map.add_car()

    def update(self):
        self.map.update()

    def draw(self, surface):
        self.map.draw(surface)

    def key_pressed(self, key):
        if key == pygame.K_ESCAPE:
            Signal().send_signal("escape")
            return True
        elif key == pygame.K_SPACE:
            Signal().send_signal("next")
            return True
        elif key == pygame.K_p:
            self.map.pause()
        return False

    def mouse_pressed(self, pos):
        self.map.mouse_pressed(pos)
